/*----------------------------------------------------------------------
  SDL Benchmark - Roberto Prieto 
  Copyright (C) 2013-2021 MegaStorm Systems
  contact@megastormsystems.com - http://www.megastormsystems.com

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

 ------------------------------------------------------------------------
                                SDL Benchmark Info
    
    - Uses SDL 1.2 Experimental with glSDL backend
    - DirectX Software, Windib and glSDL
    - 32 bits version

    Changelog:
 
 v1.20 Feb/2021 - Remove DirectX Hardware renderer.
 v1.00 Sept/2015 - First version.
   
 ------------------------------------------------------------------------ */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <windows.h>
#include <winerror.h>
#include "include/SDL.h"
#include "include/SDL_opengl.h"
#include "../SDL-Benchmark.h"

// Common vars
static SDL_Rect *positions;
static SDL_Rect *velocities;
static int sprite_w, sprite_h;
SDLBenchParameters SDLBenchParams; // Parameters struct

// SDL vars
SDL_Surface *screen;
SDL_Surface **pSpriteTable = NULL;
//#define DEBUG_RENDERING

int RemoveSprites(int iNumRealSpr)
{
    int i = 0;

    while(i < iNumRealSpr)
    {
        if(pSpriteTable != NULL)
        {
            if(pSpriteTable[i] != NULL) SDL_FreeSurface(pSpriteTable[i]);
            pSpriteTable[i] = NULL;
        }
        ++i;        
    }
    free(pSpriteTable);    
    pSpriteTable = NULL;
    return 0;
}

int LoadSprite(char *file, int iNumRealSpr)
{    
    SDL_Surface *temp;
    int i = 0;
    
    // Create a surface of each sprite
    int bFlagError = 0;
    pSpriteTable = new SDL_Surface *[iNumRealSpr];    
    while(i < iNumRealSpr)
    {
        // Load the sprite image
        temp = SDL_LoadBMP(file);
        if(temp == NULL) 
        {
            fprintf(stderr, "Couldn't load %s: %s", file, SDL_GetError());
            bFlagError = 1;
        }
        else
        {
            sprite_w = temp->w;
            sprite_h = temp->h;
            // Set transparent pixel as the pixel at (0,0)
            if(temp->format->palette ) { SDL_SetColorKey(temp, (SDL_SRCCOLORKEY|SDL_RLEACCEL),*(Uint8 *)temp->pixels); }
            else SDL_SetColorKey(temp, (SDL_SRCCOLORKEY|SDL_RLEACCEL),0);

            pSpriteTable[i] = SDL_DisplayFormat(temp);
            if(!pSpriteTable[i]) 
            {            
                fprintf(stderr, "Couldn't create texture for sprite %d: %s\n", i, SDL_GetError());
                bFlagError = 1;
            }
            SDL_FreeSurface(temp);
        }
        ++i;        
    }
    
    if(bFlagError == 1) 
    {
        RemoveSprites(iNumRealSpr);
        return -1;
    }
    return 0;
}

void MoveSprites(int iNumSpr, int iNumRealSpr)
{
    int i;
    SDL_Rect *position, *velocity;
    #ifdef DEBUG_RENDERING
     LARGE_INTEGER StartingTime, EndingTime, measure1, measure2;
     QueryPerformanceCounter(&StartingTime);
    #endif

    // Grey background
    SDL_FillRect(screen, NULL, 0xFFA0A0A0);
    #ifdef DEBUG_RENDERING
     QueryPerformanceCounter(&measure1);
    #endif

    // Move the sprite, bounce at the wall, and draw
    for(i = 0; i < iNumSpr; ++i ) 
    {
        position = &positions[i];
        velocity = &velocities[i];
        position->x += velocity->x;
        if ( (position->x < 0) || (position->x >= (screen->w - sprite_w)) ) 
        {
            velocity->x = -velocity->x;
            position->x += velocity->x;
        }
        position->y += velocity->y;
        if ( (position->y < 0) || (position->y >= (screen->h - sprite_w)) ) 
        {
            velocity->y = -velocity->y;
            position->y += velocity->y;
        }

        // Blit the sprite onto the screen
        if(iNumRealSpr == 1) SDL_BlitSurface(pSpriteTable[0], NULL, screen, position);
        else SDL_BlitSurface(pSpriteTable[i], NULL, screen, position);        
    }
    #ifdef DEBUG_RENDERING
     QueryPerformanceCounter(&measure2);
    #endif

    // Update the screen!
    SDL_Flip(screen);
    #ifdef DEBUG_RENDERING
     QueryPerformanceCounter(&EndingTime);
     printf("FrameTime: %llu + %llu + %llu = %llu\n", 
        measure1.QuadPart - StartingTime.QuadPart, 
        measure2.QuadPart - measure1.QuadPart, 
        EndingTime.QuadPart - measure2.QuadPart, 
        EndingTime.QuadPart - StartingTime.QuadPart);     
    #endif
}

int main(int argc, char *argv[])
{    
    int width, height;
    Uint8  video_bpp;
    Uint32 videoflags;
    int i, done, iMaxSprites, iRealSprites, iRenderer = 0;
    Uint32 iStart, iFrames;
    double fFPS_1, fFPS_2;   
    char szRenderer[64];

    // Print information and argument parsing
    parseAndInformation(argc, argv, SDLBENCH_SDL_VERSION" Benchmarks", &SDLBenchParams);
  
    // Max and real sprites
    iMaxSprites = SDLBenchParams.iNumSprites;
    if(iMaxSprites < SDLBenchParams.iNumSprites2) iMaxSprites = SDLBenchParams.iNumSprites2;
    if(SDLBenchParams.iFillRateMode == 0) iRealSprites = iMaxSprites;
    else iRealSprites = 1;
    width = SDLBenchParams.iVideoWidth;
    height = SDLBenchParams.iVideoHeight;
    video_bpp = 32;

    // Start with DirectX in Software mode
    i = _putenv("SDL_VIDEODRIVER=windib");
    videoflags = SDL_SWSURFACE; // | SDL_FULLSCREEN;
    strcpy(szRenderer,"WinDib");

 REINIT_SDL:
    // Initialize 
    if(SDL_Init(SDL_INIT_VIDEO) < 0 ) return -1;
    SDL_ShowCursor(0);
    printf("SDL renderer: %s\n", szRenderer);

    // Set video mode
    screen = SDL_SetVideoMode(width, height, video_bpp, videoflags);	
    if(!screen ) 
    {	
        SDL_Quit();
        return -2;	
    }
    
    // Load the sprite
    if(LoadSprite(SDLBenchParams.szIcon, iRealSprites) < 0) 
    { 
        fprintf(stderr, "Couldn't load %s: %s\n", SDLBenchParams.szIcon, SDL_GetError());
        SDL_Quit(); 
        return -3;
    }
    
    // Allocate memory for the sprite info and initialize it    
    positions = (SDL_Rect *) malloc(iMaxSprites * sizeof(SDL_Rect));
    velocities = (SDL_Rect *) malloc(iMaxSprites * sizeof(SDL_Rect));
   
    srand((unsigned int)time(NULL));
    for(i = 0; i < iMaxSprites; ++i) 
    {
        positions[i].x = rand() % (SDLBenchParams.iVideoWidth - sprite_w);
        positions[i].y = rand() % (SDLBenchParams.iVideoHeight - sprite_h);
        positions[i].w = sprite_w;
        positions[i].h = sprite_h;
        velocities[i].x = 0;
        velocities[i].y = 0;
        while(!velocities[i].x && !velocities[i].y) 
        {
            velocities[i].x = (rand() % (MAX_SPEED * 2 + 1)) - MAX_SPEED;
            velocities[i].y = (rand() % (MAX_SPEED * 2 + 1)) - MAX_SPEED;
        }
    }
    // Debug information
    //#define DEBUG_BACKEND
    #ifdef DEBUG_BACKEND
        char szRealBE[20]; char *szEnv; 
        SDL_VideoDriverName(szRealBE,20);
        szEnv = getenv("SDL_VIDEODRIVER");
        printf(" Test using '%s' engine (Real: '%s' - Requested: '%s').\n",szRenderer,szRealBE,szEnv);
        printf(" Screen is at %d bits per pixel\n",screen->format->BitsPerPixel);	
        if((screen->flags & SDL_HWSURFACE) == SDL_HWSURFACE ) printf(" Screen is in video memory\n");	
        else printf(" Screen is in system memory\n");	
        if((sprite->flags & SDL_HWSURFACE) == SDL_HWSURFACE ) printf(" Sprite is in video memory\n");	
        else printf(" Sprite is in system memory\n");	
        
        // Run a sample blit to trigger blit acceleration
        { 
            SDL_Rect dst;		
            dst.x = 0;		
            dst.y = 0;		
            dst.w = sprite->w;		
            dst.h = sprite->h;		
            SDL_BlitSurface(sprite, NULL, screen, &dst);		
            SDL_FillRect(screen, &dst, 0x000000);	
        }	
        if((sprite->flags & SDL_HWACCEL) == SDL_HWACCEL )	printf(" Sprite blit uses hardware acceleration\n");	
        if((sprite->flags & SDL_RLEACCEL) == SDL_RLEACCEL ) printf(" Sprite blit uses RLE acceleration\n");
    #endif  
       
    // ---Main render loop for iNumSprites---
    iFrames = 0;	
    iStart = SDL_GetTicks();	
    done = 0;	
    while(!done) 
    {				
        ++iFrames;
        MoveSprites(SDLBenchParams.iNumSprites, iRealSprites);        
        if(( (float)(SDL_GetTicks() - iStart) / 1000) >= SDLBenchParams.dBaseTime) { done = 1; }
    }
    // Print out some timing information
    if(SDL_GetTicks() > iStart) 
    {
        fFPS_1 = (double)iFrames / (double)((clock() - iStart) / 1000.0);
        printf(" - Moving %d sprites at %2.2f frames per second\n", SDLBenchParams.iNumSprites, fFPS_1);
    }	
    // ------

    // ---Main render loop for iNumSprites2---
    iFrames = 0;	
    iStart = SDL_GetTicks();	
    done = 0;	
    while(!done) 
    {				
        ++iFrames;
        MoveSprites(SDLBenchParams.iNumSprites2, iRealSprites);
        if(( (float)(SDL_GetTicks() - iStart) / 1000) >= SDLBenchParams.dBaseTime) { done = 1; }
    }
    // Print out some timing information
    if(SDL_GetTicks() > iStart) 
    {
        fFPS_2 = (double)iFrames / (double)((clock() - iStart) / 1000.0);
        printf(" - Moving %d sprites at %2.2f frames per second\n\n", SDLBenchParams.iNumSprites2, fFPS_2);
    }	
    // ------
    
    // Free memory buffer
    free(positions);
    free(velocities);
    RemoveSprites(iRealSprites);
    SDL_Quit();

    // Integrated mode: extended information output
    if(SDLBenchParams.iIntegrated)
    {
        FILE *fp = fopen("sdl.table","at");
        fprintf(fp,"          <tr><td align=center><b><font size=2 color=#000080>%s - %s", SDLBENCH_SDL_VERSION, szRenderer);
        fprintf(fp,"</font></b></td>");
        fprintf(fp,"<td align=center><font size=2 color=#800000>%2.2f</font></td>",fFPS_1);
        fprintf(fp,"<td align=center><font size=2 color=#800000>%2.2f</font></td></tr>\n",fFPS_2);
        fclose(fp);
        fp = fopen("SDL-Benchmark.graph","at");
        fprintf(fp,"%s (%s), %2.2f, %2.2f\n", SDLBENCH_SDL_VERSION, szRenderer, fFPS_1, fFPS_2);
        fclose(fp);
    }
    
    // Renderer selector
    switch(iRenderer)
    {
    case 0: 
        ++iRenderer; // Next renderer
        strcpy(szRenderer,"DirectX");
        i = _putenv("SDL_VIDEODRIVER=directx");
        videoflags = SDL_SWSURFACE; // | SDL_FULLSCREEN;
        goto REINIT_SDL;
        break;
    case 1: 
        ++iRenderer; // Next renderer
        strcpy(szRenderer,"glSDL");
        i = _putenv("SDL_VIDEODRIVER=glSDL");
        videoflags = SDL_HWSURFACE|SDL_DOUBLEBUF; // |SDL_FULLSCREEN
        goto REINIT_SDL;
        break;
    }             

    // End!    
    return 0;
}

